import os

import pandas as pd


def read_csv(dataset_path, table, timestamp=False):
    if timestamp:
        timestamp = ['timestamp']

    path = os.path.join(dataset_path, table + '.csv')
    return pd.read_csv(path, parse_dates=timestamp, infer_datetime_format=True)


def load_data(dataset_path, target_table, target_column, readings_table='readings',
              turbines_table='turbines', signals_table='signals'):

    tables = {
        'readings': read_csv(dataset_path, readings_table, True),
        'signals': read_csv(dataset_path, signals_table),
        'turbines': read_csv(dataset_path, turbines_table),
    }

    X = read_csv(dataset_path, target_table, True)
    y = X.pop(target_column)

    return X, y, tables



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `greenguard.pipeline` module."""
from unittest import TestCase
from unittest.mock import Mock, patch

from greenguard.pipeline import GreenGuardPipeline


class TestGreenGuardPipeline(TestCase):
    """Tests for `TimeSeriesClassifier`."""

    @patch('greenguard.pipeline.MLPipeline.from_dict')
    def test_fit(self, from_dict_mock):
        """fit prepare the pipeline to make predictions based on the given data."""

        # Setup
        pipeline_mock = Mock()
        from_dict_mock.return_value = pipeline_mock

        # Run
        instance = GreenGuardPipeline(dict(), 'accuracy')
        instance.fit('an_X', 'a_y', {'some': 'tables'})

        # Asserts
        from_dict_mock.assert_called_once_with(dict())
        assert instance._pipeline == pipeline_mock

        pipeline_mock.fit.assert_called_once_with('an_X', 'a_y', entityset=None, some='tables')

        assert instance.fitted

    @patch('greenguard.pipeline.MLPipeline.from_dict')
    def test_predict(self, from_dict_mock):
        """predict produces results using the pipeline."""
        # Setup
        pipeline_mock = Mock()
        from_dict_mock.return_value = pipeline_mock

        # Run
        instance = GreenGuardPipeline(dict(), 'accuracy')
        instance.fitted = True
        instance.predict('an_X', {'some': 'tables'})

        # Asserts
        pipeline_mock.predict.assert_called_once_with('an_X', entityset=None, some='tables')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

try:
    with open('README.md') as readme_file:
        readme = readme_file.read()
except IOError:
    readme = ''

try:
    with open('HISTORY.md') as history_file:
        history = history_file.read()
except IOError:
    history = ''

install_requires = [
    'baytune>=0.2.3,<0.3',
    'mlblocks>=0.3.0,<0.4',
    'mlprimitives>=0.1.8,<0.2',
    'numpy>=1.15.4,<1.17',
    'pymongo>=3.7.2,<4',
    'scikit-learn>=0.20.1,<0.21',
]

setup_requires = [
    'pytest-runner>=2.11.1',
]

tests_require = [
    'coverage>=4.5.1',
    'pytest>=3.4.2',
    'tox>=2.9.1'
]

development_requires = [
    # general
    'bumpversion>=0.5.3',
    'pip>=10.0.0',
    'watchdog>=0.8.3',

    # docs
    'm2r>=0.2.0',
    'Sphinx>=1.7.1',
    'sphinx_rtd_theme>=0.2.4',
    'recommonmark>=0.4.0',

    # style check
    'flake8>=3.5.0',
    'isort>=4.3.4',

    # fix style issues
    'autoflake>=1.1',
    'autopep8>=1.3.5',

    # distribute on PyPI
    'twine>=1.10.0',
    'wheel>=0.30.0',

    # Jupyter
    'jupyter>=1.0.0',
]


setup(
    author="MIT Data To AI Lab",
    author_email='dailabmit@gmail.com',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    description=(
        "The GreenGuard project is a collection of end-to-end solutions for machine"
        " learning tasks commonly found in monitoring wind energy production systems."
    ),
    extras_require={
        'test': tests_require,
        'dev': development_requires + tests_require,
    },
    include_package_data=True,
    install_requires=install_requires,
    keywords='wind machine learning green guard',
    license="MIT license",
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/markdown',
    name='greenguard',
    packages=find_packages(include=['greenguard', 'greenguard.*']),
    python_requires='>=3.5',
    setup_requires=setup_requires,
    test_suite='tests',
    tests_require=tests_require,
    url='https://github.com/D3-AI/GreenGuard',
    version='0.1.0',
    zip_safe=False,
)



# -*- coding: utf-8 -*-

import logging
import os
import shutil

import pandas as pd

LOGGER = logging.getLogger(__name__)

DATA_PATH = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    'data'
)
BUCKET = 'd3-ai-greenguard'
S3_URL = 'https://{}.s3.amazonaws.com/{}.csv'


def download(name):
    """Load the CSV with the given name from S3.

    If the CSV has never been loaded before, it will be downloaded
    from the [d3-ai-orion bucket](https://d3-ai-orion.s3.amazonaws.com)
    and then cached inside the `data` folder, within the `orion` package
    directory, and then returned.

    Otherwise, if it has been downloaded and cached before, it will be directly
    loaded from the `orion/data` folder without contacting S3.

    Args:
        name (str): Name of the CSV to load.

    Returns:
        If no test_size is given, a single pandas.DataFrame is returned containing all
        the data. If test_size is given, a tuple containing one pandas.DataFrame for
        the train split and another one for the test split is returned.
    """

    filename = os.path.join(DATA_PATH, name + '.csv')
    if os.path.exists(filename):
        data = pd.read_csv(filename)

    else:
        url = S3_URL.format(BUCKET, name)

        LOGGER.debug('Downloading CSV %s from %s', name, url)
        os.makedirs(DATA_PATH, exist_ok=True)
        data = pd.read_csv(url)
        data.to_csv(filename, index=False)

    return data


class GreenGuardLoader(object):

    def __init__(self, dataset_path, target='targets', target_column='target',
                 readings='readings', turbines='turbines', signals='signals', gzip=False):

        self._dataset_path = dataset_path
        self._target = target
        self._target_column = target_column
        self._readings = readings
        self._turbines = turbines
        self._signals = signals
        self._gzip = gzip

    def read_csv(self, table, timestamp=False):
        if timestamp:
            timestamp = ['timestamp']

        path = os.path.join(self._dataset_path, table + '.csv')
        if self._gzip:
            path += '.gz'

        return pd.read_csv(path, parse_dates=timestamp, infer_datetime_format=True)

    def load(self, target=True):
        tables = {
            'readings': self.read_csv(self._readings, True),
            'signals': self.read_csv(self._signals),
            'turbines': self.read_csv(self._turbines),
        }

        X = self.read_csv(self._target, True)
        if target:
            y = X.pop(self._target_column)
            return X, y, tables

        else:
            return X, tables


def load_demo():
    demo_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'demo')
    if os.path.exists(demo_path):
        loader = GreenGuardLoader(demo_path, gzip=True)
        return loader.load()

    else:
        os.mkdir(demo_path)
        try:
            loader = GreenGuardLoader('https://d3-ai-greenguard.s3.amazonaws.com/', gzip=True)
            X, tables = loader.load(target=False)
            X.to_csv(os.path.join(demo_path, 'targets.csv.gz'), index=False)
            for name, table in tables.items():
                table.to_csv(os.path.join(demo_path, name + '.csv.gz'), index=False)

            y = X.pop('target')
            return X, y, tables
        except Exception:
            shutil.rmtree(demo_path)
            raise



# -*- coding: utf-8 -*-

import json
import logging
import os
from collections import defaultdict

import cloudpickle
import numpy as np
from btb import HyperParameter
from btb.tuning import GP
from mlblocks import MLPipeline
from sklearn.exceptions import NotFittedError
from sklearn.metrics import accuracy_score, mean_absolute_error, mean_squared_error, r2_score
from sklearn.model_selection import KFold, StratifiedKFold

LOGGER = logging.getLogger(__name__)


PIPELINES_DIR = os.path.join(os.path.dirname(__file__), 'pipelines')

METRICS = {
    'accuracy': (accuracy_score, False),
    'r2': (r2_score, False),
    'mse': (mean_squared_error, True),
    'mae': (mean_absolute_error, True)
}


class GreenGuardPipeline(object):

    template = None
    fitted = False
    score = None

    _cv_class = None
    _metric = None
    _cost = False
    _tuner = None
    _pipeline = None

    def _get_cv(self, stratify, cv_splits, shuffle, random_state):
        if stratify:
            cv_class = StratifiedKFold
        else:
            cv_class = KFold

        return cv_class(n_splits=cv_splits, shuffle=shuffle, random_state=random_state)

    def _load_template(self, template):
        if not os.path.isfile(template):
            template = os.path.join(PIPELINES_DIR, template + '.json')

        with open(template, 'r') as template_file:
            return json.load(template_file)

    def _load_mlpipeline(self, template):
        if not isinstance(template, dict):
            template = self._load_template(template)

        self.template = template

        return MLPipeline.from_dict(template)

    def __init__(self, template, metric, cost=False, hyperparameters=None,
                 stratify=True, cv_splits=5, shuffle=True, random_state=0):

        self._cv = self._get_cv(stratify, cv_splits, shuffle, random_state)

        if isinstance(metric, str):
            metric, cost = METRICS[metric]

        self._metric = metric
        self._cost = cost

        self._pipeline = self._load_mlpipeline(template)

        if hyperparameters:
            self._pipeline.set_hyperparameters(hyperparameters)

    def get_hyperparameters(self):
        return self._pipeline.get_hyperparameters()

    def set_hyperparameters(self, hyperparameters):
        self._pipeline.set_hyperparameters(hyperparameters)
        self.fitted = False

    @staticmethod
    def _clone_pipeline(pipeline):
        return MLPipeline.from_dict(pipeline.to_dict())

    def _is_better(self, score):
        if self._cost:
            return score < self.score

        return score > self.score

    def _get_tunables(self):
        tunables = []
        tunable_keys = []
        for block_name, params in self._pipeline.get_tunable_hyperparameters().items():
            for param_name, param_details in params.items():
                key = (block_name, param_name)
                param_type = param_details['type']
                param_type = 'string' if param_type == 'str' else param_type

                if param_type == 'bool':
                    param_range = [True, False]
                else:
                    param_range = param_details.get('range') or param_details.get('values')

                value = HyperParameter(param_type, param_range)
                tunables.append((key, value))
                tunable_keys.append(key)

        return tunables, tunable_keys

    def _score_pipeline(self, pipeline, X, y, tables):
        scores = []

        for fold, (train_index, test_index) in enumerate(self._cv.split(X, y)):
            LOGGER.debug('Scoring fold %s', fold)
            X_train, X_test = X.iloc[train_index], X.iloc[test_index]
            y_train, y_test = y.iloc[train_index], y.iloc[test_index]

            pipeline = self._clone_pipeline(pipeline)
            pipeline.fit(X_train, y_train, **tables)

            predictions = pipeline.predict(X_test, **tables)
            score = self._metric(y_test, predictions)

            LOGGER.debug('Fold fold %s score: %s', fold, score)
            scores.append(score)

        return np.mean(scores)

    def _to_dicts(self, hyperparameters):

        params_tree = defaultdict(dict)
        for (block, hyperparameter), value in hyperparameters.items():
            if isinstance(value, np.integer):
                value = int(value)

            elif isinstance(value, np.floating):
                value = float(value)

            elif isinstance(value, np.ndarray):
                value = value.tolist()

            elif value == 'None':
                value = None

            params_tree[block][hyperparameter] = value

        return params_tree

    def _to_tuples(self, params_tree, tunable_keys):
        param_tuples = defaultdict(dict)
        for block_name, params in params_tree.items():
            for param, value in params.items():
                key = (block_name, param)
                if key in tunable_keys:
                    param_tuples[key] = 'None' if value is None else value

        return param_tuples

    def _get_tuner(self):
        tunables, tunable_keys = self._get_tunables()
        tuner = GP(tunables)

        # Inform the tuner about the score that the default hyperparmeters obtained
        param_tuples = self._to_tuples(self._pipeline.get_hyperparameters(), tunable_keys)
        tuner.add(param_tuples, self.score)

        return tuner

    def tune(self, X, y, tables, iterations=10):
        tables.setdefault('entityset', None)
        if not self._tuner:
            LOGGER.info('Scoring the default pipeline')
            self.score = self._score_pipeline(self._pipeline, X, y, tables)
            LOGGER.info('Default Pipeline score: %s', self.score)

            self._tuner = self._get_tuner()

        for i in range(iterations):
            LOGGER.info('Scoring pipeline %s', i + 1)

            params = self._tuner.propose(1)
            param_dicts = self._to_dicts(params)

            candidate = self._clone_pipeline(self._pipeline)
            candidate.set_hyperparameters(param_dicts)

            try:
                score = self._score_pipeline(candidate, X, y, tables)

                LOGGER.info('Pipeline %s score: %s', i + 1, score)

                if self._is_better(score):
                    self.score = score
                    self.set_hyperparameters(param_dicts)

                self._tuner.add(params, score)

            except Exception:
                failed = '\n'.join('{}: {}'.format(k, v) for k, v in params.items())
                LOGGER.exception("Caught an exception scoring pipeline %s with params:\n%s",
                                 i + 1, failed)

    def fit(self, X, y, tables):
        tables.setdefault('entityset', None)
        self._pipeline.fit(X, y, **tables)
        self.fitted = True

    def predict(self, X, tables):
        if not self.fitted:
            raise NotFittedError()

        tables.setdefault('entityset', None)
        return self._pipeline.predict(X, **tables)

    def save(self, path):
        with open(path, 'wb') as pickle_file:
            cloudpickle.dump(self, pickle_file)

    @classmethod
    def load(cls, path):
        with open(path, 'rb') as pickle_file:
            return cloudpickle.load(pickle_file)
